# Astro SDK Extras project
# (c) kol, 2023-2024

""" Custom PostgreSQL table operations """

import logging
from sqlalchemy import text, Column
from psycopg2 import sql as postgres_sql
from astro.table import BaseTable, TempTable
from astro.constants import MergeConflictStrategy
from sqlalchemy.engine.base import Connection as SqlaConnection
from sqlalchemy.dialects.postgresql.base import PGDialect
from typing import Literal, Tuple

_logger = logging.getLogger('airflow.task')

def assert_is_postgres(conn: SqlaConnection):
    """ Check whether given connection is to a Postgres database. Raises `AssertException` if not. """
    assert isinstance(conn.dialect, PGDialect), f'{conn.engine.url} is not a Postgres connection'

MergeDeleteStrategy = Literal['ignore', 'logical', 'physical']
""" Deletion strategy """

def postgres_merge_tables(
        conn: SqlaConnection,
        source_table: BaseTable | None,
        target_table: BaseTable,
        source_to_target_columns_map: dict[str, str],
        target_conflict_columns: list[str],
        source_sql: str | None = None,
        if_conflicts: MergeConflictStrategy = 'update',
        delete_strategy: MergeDeleteStrategy = 'ignore',
) -> int | Tuple[int]:
    """
    Merge the source table rows into a destination table.
    The argument `if_conflicts` allows the user to define how to handle conflicts.

    Differs from Astro-SDK `PostgresDatabase.merge_table()` as it supports pre-established
    connections (probably within transaction), custom `select` as data source and logical/physical delete.
    Also, the function will return number of rows affected.

    :param conn: Connection to execute merge on.
    :param source_table: Contains the rows to be merged to the target_table. Can be `None`.
    :param target_table: Contains the destination table in which the rows will be merged
    :param source_to_target_columns_map: Dict of target_table columns names to source_table columns names
    :param target_conflict_columns: List of cols where we expect to have a conflict while combining
    :param source_sql:   Query to get rows to be merged. Either this or `source_table` has to be provided.
    :param if_conflicts: The strategy to be applied if there are conflicts.
    """

    def identifier_args(table: BaseTable):
        schema = table.metadata.schema
        return (schema, table.name) if schema else (table.name,)

    assert_is_postgres(conn)
    assert (source_table or source_sql), f'Either `source_table` or `source_sql` must be provided'

    def construct_upsert_query() -> str:
        statement = (
            "INSERT INTO {target_table} ({target_columns}) SELECT {source_columns} FROM {source_table}"
        )

        source_columns = list(source_to_target_columns_map.keys())
        target_columns = list(source_to_target_columns_map.values())

        match if_conflicts:
            case "ignore":
                statement += " ON CONFLICT ({target_conflict_columns}) DO NOTHING"
            case "update":
                statement += " ON CONFLICT ({target_conflict_columns}) DO UPDATE SET {update_statements}"

        source_column_names = [postgres_sql.Identifier(col) for col in source_columns]
        target_column_names = [postgres_sql.Identifier(col) for col in target_columns]
        update_statements = [
            postgres_sql.SQL("{col_name}=EXCLUDED.{col_name}").format(col_name=col_name) \
                for col_name in target_column_names
        ]

        query = postgres_sql.SQL(statement).format(
            target_table=postgres_sql.Identifier(*identifier_args(target_table)),
            source_table=postgres_sql.Identifier(*identifier_args(source_table)) if source_table \
                        else postgres_sql.SQL("(" + source_sql.strip(' \n\r;') + ") q"),
            target_columns=postgres_sql.SQL(",").join(target_column_names),
            source_columns=postgres_sql.SQL(",").join(source_column_names),
            update_statements=postgres_sql.SQL(",").join(update_statements),
            target_conflict_columns=postgres_sql.SQL(",").join(
                [postgres_sql.Identifier(x) for x in target_conflict_columns]
            ),
        )
        return query

    def construct_delete_query() -> str:
        statement = (
            "DELETE FROM {target_table} WHERE {target_join_columns} NOT IN (SELECT {source_join_columns} FROM {source_table})"
        )
        join_columns = postgres_sql.SQL(",").join([postgres_sql.Identifier(x) for x in target_conflict_columns])
        if len(target_conflict_columns) > 1:
            join_columns = "(" + join_columns + ")"

        query = postgres_sql.SQL(statement).format(
            target_table=postgres_sql.Identifier(*identifier_args(target_table)),
            source_table=postgres_sql.Identifier(*identifier_args(source_table)) if source_table \
                        else postgres_sql.SQL("(" + source_sql.strip(' \n\r;') + ") q"),
            target_join_columns=join_columns,
            source_join_columns=join_columns,
        )
        return query

    def construct_logical_delete_query() -> str:
        statement = (
            'UPDATE {target_table} SET "_deleted" = current_timestamp ' \
            'WHERE "_deleted" IS NULL AND {target_join_columns} NOT IN (SELECT {source_join_columns} FROM {source_table})'
        )
        join_columns = postgres_sql.SQL(",").join([postgres_sql.Identifier(x) for x in target_conflict_columns])
        if len(target_conflict_columns) > 1:
            join_columns = "(" + join_columns + ")"

        query = postgres_sql.SQL(statement).format(
            target_table=postgres_sql.Identifier(*identifier_args(target_table)),
            source_table=postgres_sql.Identifier(*identifier_args(source_table)) if source_table \
                        else postgres_sql.SQL("(" + source_sql.strip(' \n\r;') + ") q"),
            target_join_columns=join_columns,
            source_join_columns=join_columns,
        )
        return query

    sql_list = []
    match delete_strategy:
        case 'ignore':
            sql_list.append(construct_upsert_query())
        case 'physical':
            sql_list.append(construct_upsert_query())
            sql_list.append(construct_delete_query())
        case 'logical':
            sql_list.append(construct_upsert_query())
            sql_list.append(construct_logical_delete_query())

    rowcounts = []
    for sql in sql_list:
        sql = sql.as_string(conn.connection.dbapi_connection)
        _logger.info(f'Executing {sql}')
        rowcounts.append(conn.execute(text(sql)).rowcount)

    return tuple(rowcounts) if len(rowcounts) > 1 else rowcounts[0]

def _get_pgtype_name(oid: int, conn: SqlaConnection) -> str:
    """ Returns name of Postgres type identified by OID """

    # TODO: add type cache
    return conn.execute(text(f'select pg_catalog.format_type({oid}, NULL)')).one()[0]

def postgres_infer_query_structure(sql: str, conn: SqlaConnection, infer_pk: bool = False) -> BaseTable:
    """ Reconstructs an SELECT query column structure.

    Internally, the query is executed with `LIMIT 0` addition and transform resulting cursor to SQLA-column set.

    :param sql: SELECT query
    :param conn: Connection to execute query on
    :param infer_pk: If `True`, then if the 1st table field is of `integer` type,
        it will be marked as primary key
    :returns: Astro-SDK `TempTable` with `columns` attribute assigned with inferred query structure.
    """

    assert_is_postgres(conn)

    sql = sql.strip(' \n\r;') + ' LIMIT 0'
    _logger.info(f'Executing {sql}')
    cur = conn.execute(text(sql))

    native_columns = [(x.name, x.type_code, _get_pgtype_name(x.type_code, conn)) for x in cur.cursor.description]
    _logger.debug(f'Native columns retrieved from query: {native_columns}')

    defaults = {
        'default': None,
        'notnull': None,
        'domains': None,
        'enums': None,
        'schema': None,
        'comment': None,
        'generated': None,
        'identity': None,
    }
    def get_column(pos, name, code, tp_name) -> Column:
        is_pk = False
        if infer_pk and code in (20, 21, 23) and pos == 0:
            # If 1st table field is integer, consider it as PK
            _logger.info(f'Column {name} is considered to be a primary key')
            is_pk = True

        info = conn.dialect._get_column_info(name, tp_name, **defaults)
        return Column(name=info['name'], type_=info['type'], primary_key=is_pk)

    columns = [get_column(n, name, code, tp_name) for n, (name, code, tp_name) in enumerate(native_columns)]
    table = TempTable(None, columns=columns)
    _logger.info(f'Query columns: {table.columns}')

    return table



# Astro SDK Extras project
# (c) kol, 2023-2024

__version__ = '0.1.9.2'

from .operators.session import (
    open_session, 
    close_session,
    ETLSession,
    get_current_session,
    get_session_period,
    ensure_session,
)
from .operators.table import (
    load_table,
    save_table,
    declare_tables,
    transfer_table, 
    transfer_tables,
    transfer_changed_table,
    transfer_changed_tables,
    transfer_ods_table,
    transfer_ods_tables,
    transfer_actuals_table,
    transfer_actuals_tables,
    compare_table,
    compare_tables,
    update_timed_table, 
    update_timed_tables,
    transfer_stage,
    transfer_actuals,
)
from .operators.direct import (
    run_sql_template,
    run_sql_templates,
)
from .sensors.file import (
    FileChangedSensor
)
from .sensors.auto import (
    AutoOnlyTaskSensor
)
from .utils.template import (
    get_template_file,
    get_template,
)
from .utils.utils import (
    split_table_name,
    ensure_table,
    schedule_ops,
)
from .utils.datetime_local import (
    localnow,
    datetime_is_tz_aware,
    datetime_to_local,
    datetime_to_utc,
    datetime_to_tz,
    datetime_to_naive,
    days_ago,
    months_ago
)



# Astro SDK Extras project
# (c) kol, 2023

""" Utility routines testing """

import pytest
import requests
from datetime import timedelta
from sqlalchemy import Column, Integer
from astro.sql.table import Table, Metadata

try:
    from astro_extras import split_table_name, ensure_table, localnow, days_ago
except ModuleNotFoundError:
    split_table_name = ensure_table = localnow = days_ago = None

def test_split_table_name():
    if split_table_name is None:
        return pytest.skip('Astro-extras was not installed')
    
    # Test the function with a table name that includes a schema
    table_with_schema = "schema.table"
    schema, table = split_table_name(table_with_schema)
    assert schema == "schema"
    assert table == "table"

    # Test the function with a table name that doesn't include a schema
    table_without_schema = "table"
    schema, table = split_table_name(table_without_schema)
    assert schema is None
    assert table == "table"

    # Test the function with an empty table name
    empty_table_name = ""
    with pytest.raises(ValueError):
        split_table_name(empty_table_name)

    # Test the function with a table name that starts with a dot (invalid format)
    invalid_table_name = ".table"
    schema, table = split_table_name(invalid_table_name)
    assert schema is ''
    assert table == 'table'
        
def test_ensure_table():
    if ensure_table is None:
        return pytest.skip('Astro-extras was not installed')

    # Test the function passing in None returns None

    assert ensure_table(None) is None

    # Test the function passing in a Table object returns the same Table object
    metadata = Metadata()
    test_table = Table('test', metadata, columns=[Column('id', Integer, primary_key=True)])
    assert ensure_table(test_table).__dict__ == test_table.__dict__

    # TODO: Тест не проходит, а должен. Вместо схемы в объекте Metadata возвращается None. Также добавляется public схема к названию таблицы, а не должно.
    # # Test the function passing in a string returns a Table object with proper schema and table name
    # test_table_str = 'public.test'
    # expected_table = Table(name='test', conn_id=None, metadata=Metadata(schema='public'))
    
    # assert ensure_table(test_table_str).__dict__ == expected_table.__dict__

    # Test the function TypeError is raised if an unknown type is passed in
    with pytest.raises(TypeError):
        ensure_table(12345)

def test_schedule_ops(docker_ip, docker_services, airflow_credentials):
    # Get the ID of the DAG to be tested
    dag_id = 'test-schedule-ops'
    # Form URL for getting information on tasks in the DAG
    tasks_url = f"http://localhost:18080/api/v1/dags/{dag_id}/tasks"
    # Send a GET request to the URL with authorization.
    dag_tasks_response = requests.get(tasks_url, auth=airflow_credentials)
    # Check that the request is successful
    assert dag_tasks_response.ok

    # Get JSON object with task information
    dag_tasks = dag_tasks_response.json()['tasks']
    # Check that the DAG has exactly two tasks
    assert len(dag_tasks) == 2

    # Find first task
    op1 = next((task for task in dag_tasks if task['task_id'] == 'op1'), None)
    # Find second task
    op2 = next((task for task in dag_tasks if task['task_id'] == 'op2'), None)
    # Check that both tasks were found
    assert op1 is not None and op2 is not None

    # Check that task op2 depends on task op1 (determined by downstream_task_ids)
    assert 'op2' in op1['downstream_task_ids']
    # Check that task op2 has no downstream tasks (i.e., it is the last task in the DAG)
    assert not op2['downstream_task_ids']

def test_schedule_ops_parallel(docker_ip, docker_services, airflow_credentials):
    # Get the ID of the DAG to be tested
    dag_id = 'test-schedule-ops-parallel'
    # Form URL for getting information on tasks in the DAG
    tasks_url = f"http://localhost:18080/api/v1/dags/{dag_id}/tasks"
    # Send a GET request to the URL with authorization.
    dag_tasks_response = requests.get(tasks_url, auth=airflow_credentials)
    # Check that the request is successful
    assert dag_tasks_response.ok

    # Get JSON object with task information
    dag_tasks = dag_tasks_response.json()['tasks']
    # Check that the DAG has exactly four tasks
    assert len(dag_tasks) == 4

    # Find tasks
    op1 = next((task for task in dag_tasks if task['task_id'] == 'op1'), None)
    op2 = next((task for task in dag_tasks if task['task_id'] == 'op2'), None)
    op3 = next((task for task in dag_tasks if task['task_id'] == 'op3'), None)
    op4 = next((task for task in dag_tasks if task['task_id'] == 'op4'), None)
    
    # Check that all tasks were found
    assert op1 is not None and op2 is not None and op3 is not None and op4 is not None
    # Check that task op2 depends on task op1 (determined by downstream_task_ids)
    assert 'op3' in op1['downstream_task_ids']
    # Check that task op4 depends on task op3
    assert 'op4' in op2['downstream_task_ids']
    # Check that task op2 has no downstream tasks (i.e., it is the last task in the DAG)
    assert not op3['downstream_task_ids']
    # Check that task op4 has no downstream tasks (i.e., it is the last task in the DAG)
    assert not op4['downstream_task_ids']
    
def test_datetime():
    if localnow is None:
        return pytest.skip('Astro-extras was not installed')

    dttm_now = localnow()
    assert dttm_now.tzinfo is not None

    dttm = days_ago(1)
    assert (dttm_now - dttm + timedelta(seconds=1)).days == 1



# Astro SDK Extras project
# (c) kol, 2023

""" Templates unit tests """

import pandas as pd
from confsupport import run_dag, logger

def test_run_template_list(docker_ip, docker_services, airflow_credentials, source_db):
    logger.info(f'Testing template execution')
    result = run_dag('test-run-template-list', docker_ip, docker_services, airflow_credentials)
    assert result == 'success'

    with source_db.begin() as conn:
        orig_data = pd.read_sql_table('test_table_1', conn)
        new_data = pd.read_sql_table('tmp_table_data_1', conn)
        assert new_data.shape == orig_data.shape
        new_data = pd.read_sql_table('tmp_table_data_2', conn)
        assert new_data.shape == orig_data.shape

def test_run_template_session(docker_ip, docker_services, airflow_credentials, source_db):
    logger.info(f'Testing template execution in session context')
    result = run_dag('test-run-template-session', docker_ip, docker_services, airflow_credentials)
    assert result == 'success'

    with source_db.begin() as conn:
        orig_data = pd.read_sql_table('test_table_1', conn)
        new_data = pd.read_sql_table('tmp_table_data_3', conn)
        assert new_data.shape[0] == orig_data.shape[0]
        assert 'session_id' in new_data.columns



# Astro SDK Extras project
# (c) kol, 2023

""" Table unit tests """


from datetime import datetime, timedelta

from sqlalchemy import Table
from confsupport import logger

from utils import (
    update_mod_ts_tables,
    get_table_row_count,
    create_tables_tuple
)

from asserts import (
    assert_dag_run,
    assert_table_columns_existence,
    assert_tables_existence,
    assert_views_existence,
    assert_tables_equality,
    assert_compare_table_schemas,
    assert_compare_tables_contents,
    assert_table_row_count,
    assert_tables_data_in_target
)


# Table attributes that will be ignored during tests
IGNORE_ATTR = set(['session_id', '_deleted', '_modified'])
# IDs of records that will be modified for testing
MOD_IDS = {1, 2, 3}
# Get the current date and time
DT_NOW = datetime.now()
# Calculate the date 10 days before today
DATE_10_DAYS_BEFORE = (DT_NOW.date() - timedelta(days=10))
# Dictionary containing modification dates for testing purposes
MOD_DATES = {
    # Target date and time 11 days from now
    'target': (DT_NOW + timedelta(days=11)),
    # Start date 5 days from today, formatted as "YYYY-MM-DD"
    'p_start': (DT_NOW.date() + timedelta(days=5)).strftime("%Y-%m-%d"),
    # End date 15 days from today, formatted as "YYYY-MM-DD"
    'p_end': (DT_NOW.date() + timedelta(days=15)).strftime("%Y-%m-%d")
}


def test_table_load_save(
        target_db,
        docker_ip,
        docker_services,
        airflow_credentials,
        tgt_metadata
    ):
    """ 
    Test for `load_table` and `save_table` functions.

    This test verifies the correct functionality of the `load_table` and `save_table` functions.

    Args:
    -----
        target_db: SQLAlchemy engine for the target database.
        docker_ip: IP address of the Docker host.
        docker_services: Docker services for managing containers.
        airflow_credentials: Credentials for Airflow in the form (username, password).
        tgt_metadata: SQLAlchemy MetaData object for the target database.

    Raises:
    -------
        AssertionError: If any of the assertions fail.
    """
    logger.info(f'Testing table load and save')
    table_name = 'tmp_test_table_1'
    # Ensure the DAG run named 'test-table-load_save' completes successfully
    assert_dag_run('test-table-load_save', docker_ip, docker_services, airflow_credentials)
    # Check if the temporary test table 'tmp_test_table_1' exists in the target database
    assert_tables_existence(target_db, [table_name])
    # Verify that 'tmp_test_table_1' has exactly 100 rows
    assert_table_row_count(target_db, Table(table_name, tgt_metadata, autoload_with=target_db), 100)
    # Check if the column 'some_column' exists in 'tmp_test_table_1'
    assert_table_columns_existence(target_db, table_name, ['some_column'])


def test_table_save_fail(
        docker_ip,
        docker_services,
        airflow_credentials,
        target_db
    ):
    """ 
    Test that `save_table()` function fails if no target table exists.

    This test verifies that the `save_table()` function correctly fails when attempting to save to a non-existing table.

    Args:
    -----
        docker_ip: IP address of the Docker host.
        docker_services: Docker services for managing containers.
        airflow_credentials: Credentials for Airflow in the form (username, password).
        target_db: SQLAlchemy engine for the target database.

    Raises:
    -------
        AssertionError: If any of the assertions fail.
    """
    logger.info(f'Testing table save to non-existing table')
    # Verify that the DAG run failed
    assert_dag_run('test-table-save_fail', docker_ip, docker_services, airflow_credentials, success=False)
    # Check that the non-existing table does not exist in the target database
    assert_tables_existence(target_db, ['tmp_test_table_11'], success=False)


def test_transfer_tables(
        docker_ip,
        docker_services,
        airflow_credentials,
        source_db,
        target_db,
        src_metadata,
        tgt_metadata
    ):
    """
    Test function for transfer_table() / transfer_tables() functions.

    This test verifies the correct functioning of the `transfer_table()` and `transfer_tables()` functions.
    For `test_table_2`, a template is used to select 11 rows, while for other tables, 
    records are completely reloaded. The session is not in use.

    Args:
    -----
        docker_ip (str): IP address of the Docker host.
        docker_services (pytest_docker.plugin.Services): Docker services for managing containers.
        airflow_credentials (tuple): Credentials for Airflow in the form (username, password).
        source_db (Engine): SQLAlchemy engine for the source database.
        target_db (Engine): SQLAlchemy engine for the target database.
        src_metadata (MetaData): SQLAlchemy MetaData object for the source database.
        tgt_metadata (MetaData): SQLAlchemy MetaData object for the target database.

    Returns:
    --------
        None

    Raises:
    -------
        AssertionError: If any of the assertions fail.
    """
    logger.info(f'Testing transfer_tables function')
    
    tables_name = ['test_table_1', 'test_table_2', 'test_table_3']
    
    # Check the existence of the table in the target database
    assert_tables_existence(target_db, tables_name)
    
    # Create SQLAlchemy Table objects (source)
    test_table_1_source, test_table_2_source, test_table_3_source = create_tables_tuple(
        source_db,
        tables_name,
        src_metadata
    )

    # Create SQLAlchemy Table objects (target)
    test_table_1_target, test_table_2_target, test_table_3_target = create_tables_tuple(
        target_db,
        tables_name,
        tgt_metadata
    )

    # Verify the successful execution of the DAG
    assert_dag_run('test-transfer-tables', docker_ip, docker_services, airflow_credentials)

    # Compare table schemas between the source and target databases
    assert_compare_table_schemas(test_table_2_source, test_table_2_target)

    # Verify the row count in the target table
    assert_table_row_count(target_db, test_table_2_target, 11)

    # Compare data in tables between the source and target databases, ignoring certain columns
    assert_tables_equality(
        source_db,
        target_db,
        [(test_table_1_source, test_table_1_target),
         (test_table_3_source, test_table_3_target)],
        ignore_cols=IGNORE_ATTR)


def test_transfer_tables_session(
        docker_ip,
        docker_services,
        airflow_credentials,
        source_db,
        target_db,
        src_metadata,
        tgt_metadata
    ):
    """
    Test function for the transfer_tables() operator within an ETL session.

    This test validates the transfer of data between tables in the source and target databases 
    and checks for data integrity, schema consistency, and row count accuracy. The test runs 
    the DAG twice: first to perform an initial load for a given period, and second to load 
    the updated delta for a new period. For `test_table_5`, a template is used to select a 
    specific number of records by identifiers within the period, while for other tables, 
    data selection is done for the entire period.

    Args:
    -----
        docker_ip (str): IP address of the Docker host.
        docker_services (pytest_docker.plugin.Services): Docker services for managing containers.
        airflow_credentials (tuple): Credentials for Airflow in the form (username, password).
        source_db (Engine): SQLAlchemy engine for the source database.
        target_db (Engine): SQLAlchemy engine for the target database.
        src_metadata (MetaData): SQLAlchemy MetaData object for the source database.
        tgt_metadata (MetaData): SQLAlchemy MetaData object for the target database.

    Returns:
    --------
        None

    Raises:
    -------
        AssertionError: If any of the assertions fail.
    """
    logger.info(f'Testing transfer_tables_session function under ETL session')
    
    tables_name = ['test_table_4', 'test_table_5', 'test_table_6']
    
    # Check the existence of the tables in the target database
    assert_tables_existence(target_db, tables_name)
    
    # Create SQLAlchemy Table objects (source and target)
    test_table_4_source, test_table_5_source, test_table_6_source = create_tables_tuple(
        source_db,
        tables_name,
        src_metadata
    )
    
    test_table_4_target, test_table_5_target, test_table_6_target = create_tables_tuple(
        target_db,
        tables_name,
        tgt_metadata
    )

    assert_views_existence(target_db, [f"{test_table_4_target.name}_a"])
    test_table_4_actuals = Table(f"{test_table_4_target.name}_a", tgt_metadata, autoload_with=target_db)
    
    for table in tables_name:
        assert_table_columns_existence(target_db, table, ['session_id'])
    
    # Initial load: Run the DAG and assert success
    assert_dag_run(
        'test-transfer-tables-session',
        docker_ip,
        docker_services,
        airflow_credentials,
        params = f'{{"period": "[{DATE_10_DAYS_BEFORE}, {DT_NOW.date() + timedelta(days=1)}]"}}'
    )

    # Assert that the data in 'test_table_4' and 'test_table_6' in the source and target databases are equal
    # Assert that the data in 'test_table_4' using the actuals view is equal
    assert_tables_equality(
        source_db,
        target_db,
        [(test_table_4_source, test_table_4_target),
         (test_table_6_source, test_table_6_target),
         (test_table_4_source, test_table_4_actuals)],
        ignore_cols=IGNORE_ATTR
    )
    
    # Assert that the schemas of 'test_table_5' in the source and target databases are equal
    assert_compare_table_schemas(test_table_5_source, test_table_5_target, ignore_cols=IGNORE_ATTR)
    
    # Assert the row count of 'test_table_5' in the target database
    assert_table_row_count(target_db, test_table_5_target, 11)

    # Update the 'mod_ts' field for specific tables and rows in the source database
    update_mod_ts_tables(
        source_db,
        {test_table_4_source: MOD_IDS, test_table_6_source: MOD_IDS},
        MOD_DATES['target']
    )

    # Delta load: Re-run the DAG and assert success
    assert_dag_run(
        'test-transfer-tables-session',
        docker_ip,
        docker_services,
        airflow_credentials,
        params = f'{{"period": "[{MOD_DATES["p_start"]}, {MOD_DATES["p_end"]}]"}}'
    )
    
    # Assert that the data in 'test_table_4' using the actuals view is equal after updates
    assert_compare_tables_contents(
        source_db,
        target_db,
        [(test_table_4_source, test_table_4_actuals)],
        ignore_cols=IGNORE_ATTR
    )
    
    # Assert that the data in 'test_table_4', 'test_table_5', 'test_table_6' in the source and target databases are equal
    assert_compare_tables_contents(
        source_db,
        target_db,
        [
            (test_table_4_source, test_table_4_target),
            (test_table_5_source, test_table_5_target),
            (test_table_6_source, test_table_6_target)
        ],
        ignore_cols=IGNORE_ATTR,
        success=False
    )
    
    # Assert the row count of 'test_table_5' in the target database after updates
    assert_table_row_count(target_db, test_table_5_target, 11)
    
    # Assert that the data in specific tables in the source database are present in the target database
    assert_tables_data_in_target(
        source_db,
        target_db,
        [
            (test_table_4_source, test_table_4_target),
            (test_table_6_source, test_table_6_target)
        ],
        ignore_cols=IGNORE_ATTR
    )


def test_transfer_changed_tables(
        docker_ip,
        docker_services,
        airflow_credentials,
        source_db,
        target_db,
        src_metadata,
        tgt_metadata
    ):
    """
    Test function for the transfer_changed_tables() operator.

    This test verifies the correct functioning of the `transfer_changed_tables()` function,
    ensuring that only changed records are transferred. The test involves running the DAG
    to perform an initial load, updating specific records, and running the DAG again to
    transfer only the changed records.

    Args:
    -----
        docker_ip (str): IP address of the Docker host.
        docker_services (pytest_docker.plugin.Services): Docker services for managing containers.
        airflow_credentials (tuple): Credentials for Airflow in the form (username, password).
        source_db (Engine): SQLAlchemy engine for the source database.
        target_db (Engine): SQLAlchemy engine for the target database.
        src_metadata (MetaData): SQLAlchemy MetaData object for the source database.
        tgt_metadata (MetaData): SQLAlchemy MetaData object for the target database.

    Returns:
    --------
        None

    Raises:
    -------
        AssertionError: If any of the assertions fail.
    """
    logger.info(f'Testing transfer_changed_tables() function')

    table_name = ['test_table_7']

    # Assert the existence of 'test_table_7' in the target database
    assert_tables_existence(target_db, table_name)

    # Create SQLAlchemy Table objects
    test_table_7_source, = create_tables_tuple(source_db, table_name, src_metadata)
    test_table_7_target, = create_tables_tuple(target_db, table_name, tgt_metadata)

    # Check the existence of the views in the target database
    assert_views_existence(target_db, [f"{test_table_7_target.name}_a"])

    # Create actuals view table object for 'test_table_7'
    test_table_7_actuals = Table(f"{test_table_7_target.name}_a", tgt_metadata, autoload_with=target_db)

    # Initial load: Run the DAG and assert success
    assert_dag_run(
        'test-transfer-changed-tables',
        docker_ip,
        docker_services,
        airflow_credentials,
        params=f'{{"period": "[{DATE_10_DAYS_BEFORE}, {DT_NOW.date() + timedelta(days=1)}]"}}'
    )

    # Assert that the data in 'test_table_7' in the source and target databases
    assert_tables_equality(
        source_db,
        target_db,
        [(test_table_7_source, test_table_7_actuals),
         (test_table_7_source, test_table_7_target)],
        ignore_cols=IGNORE_ATTR
    )

    # Update the 'mod_ts' field for specific rows in 'test_table_7' in the source database
    update_mod_ts_tables(
        source_db,
        {test_table_7_source: MOD_IDS},
        datetime.now() + timedelta(days=4)
    )

    # Run the DAG to transfer changed records and assert success
    assert_dag_run(
        'test-transfer-changed-tables',
        docker_ip,
        docker_services,
        airflow_credentials
    )

    # Assert that the data in 'test_table_7' using the actuals view is equal after updates
    assert_compare_tables_contents(
        source_db,
        target_db,
        [(test_table_7_source, test_table_7_actuals)],
        ignore_cols=IGNORE_ATTR
    )

    # Assert that all data from the source table 'test_table_7' is in the destination table
    assert_tables_data_in_target(
        source_db,
        target_db,
        [(test_table_7_source, test_table_7_target)],
        ignore_cols=IGNORE_ATTR
    )

    # Capture the current row count of 'test_table_7' in the target database
    row_count = get_table_row_count(target_db, test_table_7_target)

    # Run the DAG again to ensure no additional records are transferred and assert success
    assert_dag_run(
        'test-transfer-changed-tables',
        docker_ip,
        docker_services,
        airflow_credentials
    )

    # Assert that the row count of 'test_table_7' in the target database has not changed
    assert_table_row_count(target_db, test_table_7_target, row_count)


def test_transfer_ods_tables(
        docker_ip,
        docker_services,
        airflow_credentials,
        source_db,
        target_db,
        src_metadata,
        tgt_metadata
    ):
    """
    Test function for the transfer_ods_tables() operator.

    This test verifies the correct functioning of the `transfer_ods_tables()` function, ensuring that
    records are correctly transferred and updated between source and target tables. The test involves 
    multiple runs of the DAG to validate initial data transfer, data consistency, and additional 
    modifications.

    Args:
    -----
        docker_ip (str): IP address of the Docker host.
        docker_services (pytest_docker.plugin.Services): Docker services for managing containers.
        airflow_credentials (tuple): Credentials for Airflow in the form (username, password).
        source_db (Engine): SQLAlchemy engine for the source database.
        target_db (Engine): SQLAlchemy engine for the target database.
        src_metadata (MetaData): SQLAlchemy MetaData object for the source database.
        tgt_metadata (MetaData): SQLAlchemy MetaData object for the target database.

    Returns:
    --------
        None

    Raises:
    -------
        AssertionError: If any of the assertions fail.
    """
    logger.info(f'Testing transfer_ods_tables() function')

    table_name = ['test_table_8']

    # Assert the existence of 'test_table_8' in the target database
    assert_tables_existence(target_db, table_name)

    # Assert that the columns '_modified' and '_deleted' exist in 'test_table_8' in the target database
    assert_table_columns_existence(target_db, table_name[0], ['_modified', '_deleted'])

    # Create SQLAlchemy Table objects (source and target)
    test_table_8_source, = create_tables_tuple(source_db, table_name, src_metadata)
    test_table_8_target, = create_tables_tuple(target_db, table_name, tgt_metadata)

    # Check the existence of the views in the target database
    assert_views_existence(target_db, [f"{test_table_8_target.name}_a"])

    # Initial load: Run the DAG and assert success
    assert_dag_run('test-transfer-ods-tables-1', docker_ip, docker_services, airflow_credentials)

    # Create actuals view table object for 'test_table_8'
    test_table_8_a = Table(f"{test_table_8_target.name}_a", tgt_metadata, autoload_with=target_db)

    # Assert that the data in 'test_table_8' in the source and target databases are equal using the actuals view
    # Assert that the data in 'test_table_8' in the source and target databases are equal
    assert_tables_equality(
        source_db,
        target_db,
        [(test_table_8_source, test_table_8_a),
         (test_table_8_source, test_table_8_target)],
        ignore_cols=IGNORE_ATTR
    )

    # Run the DAG again and assert success
    assert_dag_run('test-transfer-ods-tables-1', docker_ip, docker_services, airflow_credentials)

    # Assert that the data in 'test_table_8' in the source and target databases are equal using the actuals view
    # Assert that the data in 'test_table_8' in the source and target databases are equal
    assert_compare_tables_contents(
        source_db,
        target_db,
        [(test_table_8_source, test_table_8_a),
         (test_table_8_source, test_table_8_target)],
        ignore_cols=IGNORE_ATTR
    )

    # Capture the current row count of 'test_table_8' in the target database
    row_count = get_table_row_count(target_db, test_table_8_target)
    
    # Run the DAG to ensure additional modifications are transferred and assert success
    assert_dag_run('test-transfer-ods-tables-2', docker_ip, docker_services, airflow_credentials)
    
    # Assert that the data in 'test_table_8' in the source and target databases are equal using the actuals view
    assert_compare_tables_contents(
        source_db,
        target_db,
        [(test_table_8_source, test_table_8_a)],
        ignore_cols=IGNORE_ATTR
    )
    
    # Assert that the row count of 'test_table_8' in the target database has increased by 22
    assert_table_row_count(target_db, test_table_8_target, row_count + 22)
    
    # Assert that all data from the source table 'test_table_8' is in the destination table
    assert_tables_data_in_target(
        source_db,
        target_db,
        [(test_table_8_source, test_table_8_target)],
        ignore_cols=IGNORE_ATTR
    )



# Astro SDK Extras project
# (c) kol, 2023-2024

""" Table operations """

import logging
import pandas as pd
from functools import cached_property
from deprecated import deprecated
from sqlalchemy import Table as SqlaTable, MetaData as SqlaMetadata
from sqlalchemy import text, Integer, BigInteger, SmallInteger
from sqlalchemy.engine.base import Connection as SqlaConnection
from sqlalchemy.exc import InvalidRequestError as SqlaInvalidRequestError, OperationalError as SqlaOperationalError

from airflow.models.dag import DagContext
from airflow.utils.context import Context
from airflow.models.xcom_arg import XComArg
from airflow.utils.task_group import TaskGroup
from airflow.exceptions import AirflowFailException
from airflow.operators.generic_transfer import GenericTransfer

from astro import sql as aql
from astro.databases import create_database
from astro.sql.table import BaseTable, Table
from astro.databases.base import BaseDatabase
from astro.airflow.datasets import kwargs_with_datasets

from .session import ETLSession, ensure_session
from ..utils.utils import ensure_table, schedule_ops, is_same_database_uri
from ..utils.template import get_template, get_template_file, get_predefined_template
from ..utils.data_compare import compare_datasets, compare_timed_dict
from ..utils.postgres_sql import postgres_merge_tables, postgres_infer_query_structure

from typing import Iterable, Mapping, Type, Tuple, Literal, Any

DEFAULT_CHUNK_SIZE: int = 10000
""" Chunk size for `pd.to_sql()` calls. Set to value lesser than `DEFAULT_CHUNK_SIZE` in Astro SDK to avoid memory overload """

StageTransferMode = Literal['normal', 'compare', 'sync']
""" Stage data transfer modes """

ActualsTransferMode = Literal['update', 'replace', 'sync']
""" Actuals data transfer modes """

class TableTransfer(GenericTransfer):
    """
    Table transfer operator to be used within `transfer_table` function.
    Implements 'bulk' data transfer without any extra conditioning.
    """

    template_fields = ("sql", "preoperator", "source_table", "destination_table")
    template_ext = (".sql", ".hql" )
    template_fields_renderers = {"sql": "sql", "preoperator": "sql"}
    ui_color = "#b4f07c"

    # Public attributes
    source_db: BaseDatabase
    """ Source database object """

    dest_db: BaseDatabase
    """ Destination database object """

    source: BaseTable
    """ Source table object as it was passed in. Property `source_table_def` reflects the same table but filled with columns structure """

    source_table: str
    """ Source table fully-qualified name """

    destination: BaseTable
    """ Destination table object as it was passed in. Property `dest_table_def` reflects the same table but filled with columns structure """

    destination_table: str
    """ Destination table fully-qualified name """

    session: ETLSession | XComArg | None
    """ ETL Session. Use `ensure_session` to cast to proper session type """

    in_memory_transfer: bool
    """ Flag to use in-memory transfer (othewise it would be Airflow's cursor-based, which is several times slower) """

    chunk_size: int
    """ Chunk size for `pd.to_sql() calls """

    def __init__(
        self,
        *,
        source_table: BaseTable,
        destination_table: BaseTable,
        session: ETLSession | None = None,
        in_memory_transfer: bool = False,
        chunk_size: int | None = DEFAULT_CHUNK_SIZE,
        **kwargs,
    ) -> None:

        # source and target databases
        self.source_db = create_database(source_table.conn_id)
        self.dest_db = create_database(destination_table.conn_id)

        # source and target AstroSDK tables
        self.source = source_table
        self.destination = destination_table

        # source and target fully-qualified table names
        # self.destination_table will be set by `super().__init__()`
        self.source_table = self.source_db.get_table_qualified_name(self.source)

        # task_id would be 'transfer-<schema>_<table>'
        task_id = kwargs.pop('task_id', f'transfer-{self.source_table.replace(".", "_")}')

        # sql is either passed in by the caller or rendered from template
        # here, original `source_table` is used to match filename case
        sql = kwargs.pop('sql', self._get_sql(source_table, self.source_db, session))

        super().__init__(task_id=task_id,
                         sql=sql,
                         destination_table=self.dest_db.get_table_qualified_name(self.destination),
                         source_conn_id=self.source.conn_id,
                         destination_conn_id=self.destination.conn_id,
                         **kwargs)

        self.session = session
        self.in_memory_transfer = in_memory_transfer
        self.chunk_size = chunk_size

        # flag to prevent multiple _pre_execute() calls
        self._pre_execute_called: bool = False

        # statistics for lineage
        self._row_count: int = 0

    def _get_sql(self, table: BaseTable, db: BaseDatabase, session: ETLSession | None = None, suffix: str | None = None) -> str:
        """ Internal - get a sql statement or template for given table """

        # Check whether a template SQL exists for given table under dags\templates\<dag_id>
        # Actual query will be loaded by Airflow templating itself
        full_name = db.get_table_qualified_name(table)
        self.log.info(f'Looking up a template file for table {full_name}')
        if (sql_file := get_template_file(full_name, '.sql')) or (sql_file := get_template_file(table.name, '.sql')) :
            self.log.info(f'Using template file {sql_file}')
            return sql_file

        # Nope, load an SQL from package resources substituting template fields manually
        # SQL file names are fixed according to whether we do run under ETL session or not
        template_name = 'table_transfer_nosess.sql' if not session else 'table_transfer_sess.sql'
        template = get_predefined_template(template_name)
        sql = template.render(source_table=full_name + (suffix or ''))
        self.log.info(f'Using predefined SQL template {template_name} rendered as {sql}')
        return sql

    def _pre_execute(self, context: Context):
        """ Internal - run before execution """

        if self._pre_execute_called:
            return

        # Set debug level logging if requested upon DAG start
        # Config is: {..., "debug": true}
        assert 'dag_run' in context
        if context['dag_run'].conf.get('debug'):
            self.log.setLevel(logging.DEBUG)
            logging.getLogger('airflow.task').setLevel(logging.DEBUG)
            self.log.debug('Log level set to DEBUG')

        # Transform XComArg session into ETLSession
        self.session = ensure_session(self.session, context)
        if self.session:
            if not self.source_conn_id:
                self.source_conn_id = self.session.source_conn_id
            if not self.destination_conn_id:
                self.destination_conn_id = self.session.destination_conn_id

        if not self.source_conn_id:
            raise AirflowFailException('source connection not specified')
        if not self.destination_conn_id:
            raise AirflowFailException('destination connection not specified')
        if self.source_conn_id == self.destination_conn_id and self.source_table == self.destination_table:
            raise AirflowFailException('Source and destination must not be the same')

        self._pre_execute_called = True

    def _load_table_def(self, base_table: BaseTable, db: BaseDatabase) -> BaseTable:
        """ Internal - load single table definition """
        # Use base_table's SQLA metadata to retrieve table columns from database
        # There's something that looks like a bug in SQLA - 
        # only lower-case schema/table names could be used with reflect
        try:
            full_name = db.get_table_qualified_name(base_table).lower()
            meta = SqlaMetadata(schema=base_table.metadata.schema.lower() \
                                if base_table.metadata and base_table.metadata.schema else None)
            meta.reflect(bind=db.connection, only=[base_table.name.lower()])
        except SqlaInvalidRequestError as ex:
            self.log.error(f'Could not load table {full_name} structure on {db.conn_id}')
            return base_table

        if (sqla_table := meta.tables.get(full_name, None)) is None:
            self.log.error(f'Could not find table {full_name} on {db.conn_id}')
            return base_table

        # Construct new table object containing all props of `base_table` and additional columns metadata
        result_table = Table(
            name=base_table.name,
            conn_id=db.conn_id,
            metadata=base_table.metadata,
            temp=base_table.temp,
            columns=sqla_table.columns)
        
        self.log.debug(f'Table {full_name} columns: {",".join([f"{c.name} {c.type}" for c in result_table.columns])}')
        return result_table

    @cached_property
    def source_table_def(self) -> BaseTable:
        """ Source table object populated with metadata """
        return self._load_table_def(self.source, self.source_db)

    @cached_property
    def dest_table_def(self) -> BaseTable:
        """ Destination table object populated with metadata """
        return self._load_table_def(self.destination, self.dest_db)

    def _adjust_dtypes(self, data: pd.DataFrame, table: BaseTable) -> pd.DataFrame:
        """ Checks whether `data` column dtypes match given `table` ones and reset them to proper type.
        This could happen if, for example, source has nulls in `int` columns which would result 
        in `float` type been assigned by Pandas.
        """
        for col in table.columns:
            if col.name not in data.columns:
                self.log.warning(f'Column {col.name} of table {table.name} was not found in dataset')
                continue

            match col.type:
                case Integer() | SmallInteger() | BigInteger() if data.dtypes[col.name] == 'float':
                    self.log.info(f'Coersing column {col.name} dtype from `float` to `int`')
                    data[col.name] = data[col.name].astype('Int64')

                case _:
                    pass

        return data

    def execute(self, context: Context):
        """ Execute operator """
        self._pre_execute(context)
        if not self.in_memory_transfer:
            return super().execute(context)
        
        # upload source data to memory
        self.log.info(f'In-memory transfer using {self.sql}')
        data = pd.read_sql(self.sql, self.source_db.connection)
        if data is None or data.empty:
            self.log.info('No data to transfer')
            return

        # Adjust dtypes
        data = self._adjust_dtypes(data, self.dest_table_def)

        with self.dest_db.connection as conn, conn.begin():
            # run preoperator
            if self.preoperator:
                self.log.info(f'Executing: {self.preoperator}')
                conn.execute(text(self.preoperator))

            # save to target table
            self._row_count = len(data)
            self.log.info(f'{self._row_count} records to be transferred (chunk size is {self.chunk_size})')
            data.to_sql(
                self.destination.name,
                con=conn,
                schema=self.destination.metadata.schema if self.destination.metadata else None,
                if_exists='append',
                method='multi',
                chunksize=self.chunk_size,
                index=False,
            )

    def _make_openlineage_facets(self, task_instance):
        """ Make OpenLineage facets and dataset info """

        from openlineage.client.run import Dataset
        from openlineage.client.facet import (
            BaseFacet,
            DataSourceDatasetFacet,
            OutputStatisticsOutputDatasetFacet,
            SchemaDatasetFacet,
            SchemaField,
            SqlJobFacet,
        )

        def _make_dataset(table_def: BaseTable, db: BaseDatabase) -> Dataset:
            ns = db.openlineage_dataset_namespace()
            full_name = db.get_table_qualified_name(table_def)
            if (dbname := table_def.metadata.database or db.default_metadata.database):
                full_name = f'{dbname}.{full_name}'

            return Dataset(
                namespace=ns,
                name=f'{full_name}',
                facets={
                    "schema": SchemaDatasetFacet(
                        fields=[SchemaField(name=col.name, type=str(col.type)) for col in table_def.columns],
                    ),
                    "dataSource": DataSourceDatasetFacet(
                        name=full_name, uri=f"{ns}/{full_name}"
                    ),
                })

        # Construct input dataset...
        if self.source_table_def.openlineage_emit_temp_table_event():
            source_datasets = [_make_dataset(self.source_table_def, self.source_db)]
        else:
            source_datasets = []

        # ... output dataset ...
        if self.dest_table_def.openlineage_emit_temp_table_event():
            dest_datasets = [_make_dataset(self.dest_table_def, self.dest_db)]
        else:
            dest_datasets = []

        # ... runtime facets ...
        run_facets: dict[str, BaseFacet] = {
            "outputStatistics": OutputStatisticsOutputDatasetFacet(
                rowCount=self._row_count,
            ),
        }
        job_facets: dict[str, BaseFacet] = {
            "sql": SqlJobFacet(query=str(self.sql)),
        }
        return source_datasets, dest_datasets, run_facets, job_facets

    def get_openlineage_facets_on_complete(self, task_instance):  # skipcq: PYL-W0613
        """ Provide OpenLineage information """
        from airflow.providers.openlineage.extractors.base import OperatorLineage

        inputs, outputs, run_facets, job_facets = self._make_openlineage_facets(task_instance)
        return OperatorLineage(
            inputs=inputs, outputs=outputs, run_facets=run_facets, job_facets=job_facets
        )

class CompareTableTransfer(TableTransfer):

    template_fields = ("sql", "preoperator", "source_table", "destination_table", "destination_sql")
    template_fields_renderers = {"sql": "sql", "preoperator": "sql", "destination_sql": "sql"}
    ui_color = "#95f07c"

    def __init__(
        self,
        *,
        source_table: BaseTable,
        destination_table: BaseTable,
        session: ETLSession | None = None,
        **kwargs,
    ) -> None:

        dest_db = create_database(destination_table.conn_id, destination_table)
        dest_sql = kwargs.pop('destination_sql', self._get_sql(destination_table, dest_db, suffix='_a'))

        super().__init__(source_table=source_table,
                         destination_table=destination_table,
                         session=session,
                         **kwargs)

        self.destination_sql: str = dest_sql

    def _compare_datasets(self, src_conn: SqlaConnection, dest_conn: SqlaConnection, stop_on_first_diff: bool, logger: logging.Logger | None = None):
        """ Internal - compare source and target dictionaries """

        logger = logger or self.log

        logger.info(f'Executing: {self.sql}')
        df_src = pd.read_sql(self.sql, src_conn)
        logger.info(f'{len(df_src)} records selected on source')

        logger.info(f'Executing: {self.destination_sql}')
        df_trg  = pd.read_sql(self.destination_sql, dest_conn)
        logger.info(f'{len(df_trg)} records selected on target')

        return compare_datasets(df_src, df_trg, stop_on_first_diff=stop_on_first_diff, logger=logger)

    def execute(self, context: Context):
        """ Execute operator """
        self._pre_execute(context)
        if not self._compare_datasets(self.source_db.connection, self.dest_db.connection, stop_on_first_diff=True):
            return super().execute(context)

class SyncTableTransfer(CompareTableTransfer):
    ui_color = "#78f07c"

    def _save_data(self, data: pd.DataFrame, conn: SqlaConnection, modified: pd.Timestamp, deleted: pd.Timestamp | None, category: str) -> None:
        if data is not None and not data.empty:
            if self.session is not None:
                data.drop(columns=set(['session_id', '_modified', '_deleted']) & set(data.columns), inplace=True)
                data.insert(0, '_deleted', deleted)
                data.insert(0, '_modified', modified)
                data.insert(0, 'session_id', self.session.session_id)
            else:
                data['_modified'] = modified
                data['_deleted'] = deleted

            self.log.info(f'Saving {data.shape[0]} {category} records to {self.destination_table} (chunk size is {self.chunk_size})')
            data.to_sql(
                self.destination.name,
                con=conn,
                schema=self.destination.metadata.schema if self.destination.metadata else None,
                if_exists='append',
                method='multi',
                chunksize=self.chunk_size,
                index=False,
            )
            self._row_count += len(data)

    def execute(self, context: Context):
        """ Execute operator """

        # All the checks and strcture load
        self._pre_execute(context)

        # Verify target table structure
        if self.session is not None:
            for req_col in ['session_id', '_modified', '_deleted']:
                cols = [col.name for col in self.dest_table_def.columns if col.name.lower() == req_col]
                if not cols:
                    raise AirflowFailException(f'Invalid ODS table {self.destination_table} structure: {req_col} attribute is missing')

        # compare datasets and save delta frames to target (new/modified/deleted)
        self._row_count = 0
        with self.source_db.connection as src_conn, self.dest_db.connection as dest_conn:
            dfn, dfm, dfd = self._compare_datasets(src_conn, dest_conn, stop_on_first_diff=False)
            with dest_conn.begin():
                self._save_data(dfn, dest_conn, pd.Timestamp.utcnow(), pd.NaT, 'new')
                self._save_data(dfm, dest_conn, pd.Timestamp.utcnow(), pd.NaT, 'modified')
                self._save_data(dfd, dest_conn, pd.Timestamp.utcnow(), pd.Timestamp.utcnow(), 'deleted')

class ActualsTableTransfer(TableTransfer):

    ui_color = "#5af07d"

    def __init__(
        self,
        *,
        source_table: BaseTable,
        destination_table: BaseTable,
        session: ETLSession | None = None,
        keep_temp_table: bool = False,
        mode: ActualsTransferMode = 'update',
        **kwargs,
    ) -> None:
        self.mode: ActualsTransferMode = mode
        self.keep_temp_table: bool = keep_temp_table
        
        super().__init__(source_table=source_table, destination_table=destination_table, session=session, **kwargs)

    def _get_sql(self, table: BaseTable, db: BaseDatabase, session: ETLSession | None = None, suffix: str | None = None) -> str:
        """ Internal - get a sql statement or template for given table """

        # Find table-specific template
        full_name = db.get_table_qualified_name(table)
        if (sql := get_template(full_name, '.sql', fail_if_not_found=False)):
            # If such template exist, use it to extract data
            return sql

        # Get template SQL from package resources
        template = get_predefined_template('table_actuals_select_delta.sql')
        return template.render(source_table=full_name)

    def execute(self, context: Context):
        """ Execute operator """

        # All the checks
        self._pre_execute(context)

        # Verify structure
        if self.mode == 'sync':
            # Check source and target meet the sync requirements
            if not [c for c in self.dest_table_def.columns if c.name == '_deleted']:
                raise AirflowFailException(f'Target table {self.destination_table} does not have `_deleted` column required for mode=="sync"')

        # Build a source-to-target column mapping
        # Options:
        #   - source metadata available, target is not: target table does not exist, fail
        #   - both source and target table metadata (columns) available: build mapping the normal way
        #   - source metadata not available, target is: could happen if a custom query is used -> run query with `limit 1` addition and
        #       reconstruct structure from result
        def build_mapping(source_table_def: BaseTable, dest_table_def: BaseTable) -> Tuple[Mapping, Iterable]:
            _col_map = {}
            _id_cols = []

            for src_col in source_table_def.columns:
                dest_cols = [c for c in dest_table_def.columns if c.name.lower() == src_col.name.lower()]
                if not dest_cols:
                    self.log.warning(f'Column {src_col.name} was not found in {self.destination_table} table')
                else:
                    # If a column is PK, store it to use in `on conflict` statement part
                    _col_map[src_col.name] = dest_cols[0].name
                    if dest_cols[0].primary_key:
                        _id_cols.append(dest_cols[0].name)

            return _col_map, _id_cols

        match (bool(self.source_table_def.columns), bool(self.dest_table_def.columns)):
            case (True, False) | (False, False):
                raise AirflowFailException(f'Could not get destination table {self.destination_table} metadata, does it really exist?')

            case (True, True):
                self.log.info('Both source and destination metadata are available')
                src_tbl = self.source_table_def

            case (False, True):
                self.log.info('Source metadata is not available, attempting to retrieve from cursor')
                with self.source_db.connection as src_conn:
                    src_tbl = postgres_infer_query_structure(self.sql, src_conn, infer_pk=False)

        col_map, id_cols = build_mapping(src_tbl, self.dest_table_def)

        # Check there are any ID columns on target
        self.log.debug(f'Column mapping: {col_map}')
        self.log.debug(f'ID columns: {id_cols}')
        if not id_cols:
            raise AirflowFailException(f'Could not detect primary key on {self.destination_table}')

        # Replace `t.id` with ids and `t.*` with selection columns list and proper session_id
        id_cols_str = ", ".join([f't.{c}' for c in id_cols])
        all_cols_str = f'{self.session.session_id} as session_id, ' + ", ".join([f't.{c}' for c in col_map if c != 'session_id'])
        sql = self.sql.replace('t.id', id_cols_str).replace('t.*', all_cols_str)
        self.log.info(f'Source extraction SQL:\n{sql}')

        # Check whether the hooks points to the same database
        same_db = is_same_database_uri(self.source_db.hook.get_uri(), self.dest_db.hook.get_uri())

        # Update/merge
        with self.source_db.connection as src_conn, self.dest_db.connection as dest_conn:
            if same_db:
                # Source and destination tables are in the same database, 
                self.log.info(f'Source and destination tables are in the same database')
                with dest_conn.begin():
                    # Purge target table if requested
                    if self.mode == 'replace':
                        self.log.info(f'Purging destination table {self.destination_table}')
                        dest_conn.execute(text(f'delete from {self.destination_table}').execution_options(autocommit=True))

                    # Merge into target table directly using source sql
                    self._row_count = postgres_merge_tables(
                        conn=dest_conn,
                        source_table=None,
                        target_table=self.destination,
                        source_to_target_columns_map=col_map,
                        target_conflict_columns=id_cols,
                        source_sql=sql,
                        if_conflicts='update',
                        delete_strategy='logical' if self.mode == 'sync' else 'ignore'
                    )
                    self.log.info(f'{self._row_count} records transferred')

            else:
                # Source and destination tables are in different databases
                temp_table = Table(metadata=self.destination.metadata, temp=True)
                self.log.info(f'Source and destination tables are in different databases, transferring via temporary table')

                # Load data from source table
                data = pd.read_sql(sql, src_conn)
                if data is None or data.empty:
                    self.log.info('No data to transfer')
                    return
                self.log.info(f'{len(data)} records to be transferred (chunk size is {self.chunk_size})')

                # Adjust data types to target table structure                
                data = self._adjust_dtypes(data, self.dest_table_def)

                # Temp table has to be created 1st, otherwise column types mismatch might occur
                temp_sqla_table = SqlaTable(temp_table.name, temp_table.sqlalchemy_metadata, *([c.copy() for c in src_tbl.columns]))
                self.destination.sqlalchemy_metadata.create_all(bind=dest_conn, tables=[temp_sqla_table], checkfirst=False)
                self.log.info(f'{temp_table.name} temporary table created')

                try:
                    # Save data to temporary table
                    data.to_sql(
                        temp_table.name,
                        con=dest_conn,
                        schema=temp_table.metadata.schema if temp_table.metadata else None,
                        if_exists='append',
                        method='multi',
                        chunksize=self.chunk_size,
                        index=False,
                    )
                    self.log.info(f'Source data transferred to {temp_table.name}')

                    with dest_conn.begin() as tran:
                        # Purge target table if requested
                        if self.mode == 'replace':
                            self.log.info(f'Purging destination table {self.destination_table}')
                            dest_conn.execute(text(f'delete from {self.destination_table}'))

                        # Merge temporary and target tables
                        self.log.info(f'Merging from {self.dest_db.get_table_qualified_name(temp_table)} to {self.destination_table}')
                        self._row_count = postgres_merge_tables(
                            conn=dest_conn,
                            source_table=temp_table,
                            target_table=self.destination,
                            source_to_target_columns_map=col_map,
                            target_conflict_columns=id_cols,
                            if_conflicts='update',
                            delete_strategy='logical' if self.mode == 'sync' else 'ignore'
                        )
                        self.log.info(f'{self._row_count} records transferred')
                        tran.commit()

                finally:
                    if not self.keep_temp_table:
                        try:
                            self.dest_db.drop_table(temp_table)
                        except SqlaOperationalError as ex:
                            self.log.error(f'Cannot remove table {temp_table} due to error {ex}. Wait till DAG finishes and remove it yourself.')

class TimedTableTransfer(CompareTableTransfer):
    """
    Table transfer operator to be used within `update_timed_dict` function.
    Implements timed dictionary update behaviour.
    Requiures `xxx_a` view to exist both on source and target.
    """
    ui_color = "#3cf07e"

    def execute(self, context: Context):
        """ Execute operator """

        # All the checks
        self._pre_execute(context)

        # Get source and target data
        with self.source_db.connection as src_conn, self.dest_db.connection as dest_conn, dest_conn.begin():
            self.log.info(f'Executing: {self.sql}')
            df_src = pd.read_sql(self.sql, src_conn)
            self.log.info(f'Executing: {self.destination_sql}')
            df_trg = pd.read_sql(self.destination_sql, dest_conn)

            # Get the deltas
            df_opening, df_closing = compare_timed_dict(df_src, df_trg)
            if (df_opening is None or df_opening.empty) and (df_closing is None or df_closing.empty):
                self.log.info('No changes detected, nothing to do')
                return

            # Update records to be closed
            if df_closing is not None and not df_closing.empty:
                update_sql_template = get_predefined_template('table_timed_update.sql')
                update_sql = update_sql_template.render(destination_table=self.destination_table)
                update_params = df_closing.to_dict(orient='list')
                dest_conn.execute(text(update_sql).execution_options(autocommit=True), update_params)

            # Insert new records
            if df_opening is not None and not df_opening.empty:
                df_opening.to_sql(
                    self.destination.name,
                    con=dest_conn,
                    schema=self.destination.metadata.schema if self.destination.metadata else None,
                    if_exists='append',
                    method='multi',
                    chunksize=self.chunk_size,
                    index=False,
                )


class CompareTableOperator(CompareTableTransfer):
    """
    Table comparsion operator to be used within `compare_table` function.
    Compares source and target data and prints results to log.
    """
    ui_color = "#64bf62"

    def execute(self, context: Context):
        """ Execute operator """
        logger = logging.getLogger(f'compare_tables_logger')
        logger.setLevel(logging.DEBUG)
        if not self._compare_datasets(self.source_db.connection, self.dest_db.connection, stop_on_first_diff=True, logger=logger):
            raise AirflowFailException(f'Differences detected')

def load_table(
        table: str | BaseTable,
        conn_id: str | None = None,
        session: ETLSession | None = None,
        sql: str | None = None) -> XComArg:
    """ Loads table into memory.

    This is a wrapper over Astro-SDK `run_raw_sql` to
    load data from given database table into XCom and make it available for 
    further processing.

    SQL templating is supported, e.g. if a template for given table was found, it
    will be executed to get the data (see `astro_extras.utils.template.get_template_file`).

    Please note that in order to operate even on modest volumes of data,
    intermediate XCom storage might be required. Easiest way to set it up is to use
    local Parquet file storage by setting
    `AIRFLOW__ASTRO_SDK__XCOM_STORAGE_CONN_ID=local` environment
    variable. However, this will add extra serialization/deserialization
    operation to every task thus increasing overall DAG execution time.

    See https://astro-sdk-python.readthedocs.io/en/1.2.0/guides/xcom_backend.html
    for details.

    Args:
        table:  Either a table name or Astro-SDK `Table` object to load data from
        conn_id:    Airflow connection ID to underlying database. If not specified,
            and `Table` object is passed it, its `conn_id` attribute will be used.
        session:    `astro_extras.operators.session.ETLSession` object. 
            Used only to link up to the `open_SESSIONion` operator.
        sql:    Custom SQL to load data, used only if no SQL template found.
            If neither SQL nor template is given, all table data will be loaded.

    Results:
        `XComArg` object suitable for further manipulations with Astro-SDK functions

    Examples:
        >>> @aql.dataframe
        >>> def modify_data(data: pd.DataFrame):
        >>>     data['some_column'] = 'new_value'
        >>>     return data
        >>> data = load_table('test_table', conn_id='source_db')
        >>> modified_data = modify_data(data)
        >>> save_table(modified_data, conn_id='target_db')
    """

    if not isinstance(table, BaseTable):
        dag = DagContext.get_current_dag()
        @aql.run_raw_sql(handler=lambda result: result.fetchall(),
                        conn_id=conn_id,
                        task_id=f'load-{table}',
                        results_format='pandas_dataframe')
        def _load_table_by_name(table: str, session: ETLSession | None):
            sql_from_template = get_template(table, '.sql', dag=dag)
            return sql or sql_from_template or f'select * from {table}'

        return _load_table_by_name(table, session)
    else:
        dag = DagContext.get_current_dag()
        @aql.run_raw_sql(handler=lambda result: result.fetchall(),
                        conn_id=conn_id,
                        task_id=f'load-{table.name}',
                        results_format='pandas_dataframe')
        def _load_table(table: BaseTable, session: ETLSession | None):
            sql_file = get_template_file(table.name, '.sql', dag=dag)
            return sql or sql_file or '''select * from {{table}}'''

        return _load_table(table, session)

def save_table(
        data: XComArg,
        table: str | BaseTable,
        conn_id: str | None = None,
        session: ETLSession | None = None,
        fail_if_not_exist: bool | None = True) -> XComArg:
    """ Saves a table into database """

    table = ensure_table(table, conn_id)
    task_id = f'save-{table.name}'
    conn_id = conn_id or table.conn_id

    @aql.dataframe(if_exists='append', conn_id=conn_id, task_id=task_id)
    def _save_data(data: pd.DataFrame, session: ETLSession):
        if fail_if_not_exist:
            db = create_database(conn_id, table)
            if not db.table_exists(table):
                raise AirflowFailException(f'Table {table.name} was not found under {conn_id} connection')
        session = ensure_session(session)
        if session and 'session_id' not in data.columns:
            data.insert(0, 'session_id', session.session_id)
        return data

    return _save_data(data, session, output_table=table)

def declare_tables(
        table_names: Iterable[str],
        conn_id: str | None = None,
        schema: str | None = None,
        database: str | None = None,
        lowercase: bool = False
) -> Iterable[BaseTable]:
    """ Convert list of string table names to list of `Table` objects """
    if not lowercase:
        return [ensure_table(t, conn_id, schema, database) for t in table_names]
    else:
        return [ensure_table(t.lower(), conn_id, schema, database) for t in table_names]

@deprecated('`transfer_table()` is deprecated since 0.1.9, use `transfer_stage(mode="normal")` instead')
def transfer_table(
        source: str | BaseTable,
        target: str | BaseTable | None = None,
        source_conn_id: str | None = None,
        destination_conn_id: str | None = None,
        session: XComArg | ETLSession | None = None,
        changes_only: bool | None = None,
        **kwargs) -> XComArg:
    
    """ Cross-database data transfer (deprecated, use `transfer_stage(mode=“normal“)`). """

    return transfer_stage(
        source_tables=source, 
        target_tables=target,
        source_conn_id=source_conn_id,
        destination_conn_id=destination_conn_id,
        session=session,
        mode='normal' if not changes_only else 'compare',
        **kwargs)

@deprecated('`transfer_tables()` is deprecated since 0.1.9, use `transfer_stage(mode="normal")` instead')
def transfer_tables(
        source_tables: Iterable[str | Table],
        target_tables: Iterable[str | Table] | None = None,
        source_conn_id: str | None = None,
        destination_conn_id: str | None = None,
        group_id: str | None = None,
        num_parallel: int = 1,
        session: XComArg | ETLSession | None = None,
        changes_only: bool | None = None,
        **kwargs) -> TaskGroup:

    """ Transfer multiple tables (deprecated, use `transfer_stage(mode=“normal“)`). """

    return transfer_stage(
        source_tables=source_tables, 
        target_tables=target_tables,
        source_conn_id=source_conn_id,
        destination_conn_id=destination_conn_id,
        group_id=group_id,
        num_parallel=num_parallel,
        session=session,
        mode='normal' if not changes_only else 'compare',
        **kwargs)

@deprecated('`transfer_changed_table()` is deprecated since 0.1.9, use `transfer_stage(mode="compare")` instead')
def transfer_changed_table(
        source: str | BaseTable,
        target: str | BaseTable | None = None,
        source_conn_id: str | None = None,
        destination_conn_id: str | None = None,
        session: XComArg | ETLSession | None = None,
        **kwargs) -> XComArg:
    
    """ Transfer changed table (deprecated, use `transfer_stage(mode="compare)`). """

    return transfer_stage(
        source_tables=source, 
        target_tables=target,
        source_conn_id=source_conn_id,
        destination_conn_id=destination_conn_id,
        session=session,
        mode='compare',
        **kwargs)

@deprecated('`transfer_changed_tables()` is deprecated since 0.1.9, use `transfer_stage(mode="compare")` instead')
def transfer_changed_tables(
        source_tables: Iterable[str | Table],
        target_tables: Iterable[str | Table] | None = None,
        source_conn_id: str | None = None,
        destination_conn_id: str | None = None,
        group_id: str | None = None,
        num_parallel: int = 1,
        session: XComArg | ETLSession | None = None,
        **kwargs) -> TaskGroup:

    """ Transfer changed tables (deprecated, use `transfer_stage(mode="compare)`). """

    return transfer_stage(
        source_tables=source_tables, 
        target_tables=target_tables,
        source_conn_id=source_conn_id,
        destination_conn_id=destination_conn_id,
        group_id=group_id,
        num_parallel=num_parallel,
        session=session,
        mode='compare',
        **kwargs)

@deprecated('`transfer_ods_table()` is deprecated since 0.1.9, use `transfer_stage(mode="sync")` instead')
def transfer_ods_table(
        source: str | BaseTable,
        target: str | BaseTable | None = None,
        source_conn_id: str | None = None,
        destination_conn_id: str | None = None,
        session: XComArg | ETLSession | None = None,
        **kwargs) -> XComArg:
    
    """ Cross-database ODS-like data transfer (deprecated, use `transfer_stage(mode="sync“)`). """

    return transfer_stage(
        source_tables=source, 
        target_tables=target,
        source_conn_id=source_conn_id,
        destination_conn_id=destination_conn_id,
        session=session,
        mode='sync',
        **kwargs)

@deprecated('`transfer_ods_tables()` is deprecated since 0.1.9, use `transfer_stage(mode="sync")` instead')
def transfer_ods_tables(
        source_tables: Iterable[str | Table],
        target_tables: Iterable[str | Table] | None = None,
        source_conn_id: str | None = None,
        destination_conn_id: str | None = None,
        group_id: str | None = None,
        num_parallel: int = 1,
        session: XComArg | ETLSession | None = None,
        **kwargs) -> TaskGroup:

    """ Cross-database ODS-like data transfer (deprecated, use `transfer_stage(mode="sync“)`). """

    return transfer_stage(
        source_tables=source_tables, 
        target_tables=target_tables,
        source_conn_id=source_conn_id,
        destination_conn_id=destination_conn_id,
        group_id=group_id,
        num_parallel=num_parallel,
        session=session,
        mode='sync',
        **kwargs)

_STAGE_MODE_MAP = {
    'normal': {'op_cls': TableTransfer, 'default_group': 'transfer-tables' },
    'compare': {'op_cls': CompareTableTransfer, 'default_group': 'transfer-dict' },
    'sync': {'op_cls': SyncTableTransfer, 'default_group': 'transfer-ods-tables' },
}

def transfer_stage(
        source_tables: Table | str | Iterable[str | Table],
        target_tables: Table | str | Iterable[str | Table] | None = None,
        source_conn_id: str | None = None,
        destination_conn_id: str | None = None,
        group_id: str | None = None,
        num_parallel: int = 1,
        session: XComArg | ETLSession | None = None,
        mode: StageTransferMode = 'normal',
        **kwargs) -> XComArg | TaskGroup:

    """ Snapshot data transfer """

    if (mode_info := _STAGE_MODE_MAP.get(mode)) is None:
        raise AirflowFailException(f'Invalid stage transfer mode {mode}')

    if isinstance(source_tables, (Table, str)):
        # Single table
        source_table = ensure_table(source_tables, source_conn_id)
        dest_table = ensure_table(target_tables or source_tables, destination_conn_id)

        op = mode_info['op_cls'](
            source_table=source_table,
            destination_table=dest_table,
            session=session,
            inlets=[source_table],
            outlets=[dest_table],
            **kwargs
        )
        return XComArg(op)

    # Multiple tables
    if target_tables and len(target_tables) != len(source_tables):
        raise AirflowFailException(f'Source and target tables list size must be equal')

    target_tables = target_tables or source_tables
    group_id = group_id or mode_info['default_group']

    with TaskGroup(group_id, add_suffix_on_collision=True, prefix_group_id=False) as tg:
        ops_list = []
        for (source, target) in zip(source_tables, target_tables):
            source_table = ensure_table(source, source_conn_id)
            dest_table = ensure_table(target, destination_conn_id)
            op = mode_info['op_cls'](
                source_table=source_table,
                destination_table=dest_table,
                session=session,
                inlets=[source_table],
                outlets=[dest_table],
                **kwargs
            )
            ops_list.append(op)

        schedule_ops(ops_list, num_parallel)

    return tg

@deprecated('`transfer_actuals_table()` is deprecated since 0.1.9, use `transfer_actuals()` instead')
def transfer_actuals_table(
        source: str | BaseTable,
        target: str | BaseTable | None = None,
        source_conn_id: str | None = None,
        destination_conn_id: str | None = None,
        session: XComArg | ETLSession = None,
        transfer_delta: bool = True,
        as_ods: bool = False,
        keep_temp_table: bool = False,
        replace_data: bool = False,
        **kwargs) -> XComArg:
    
    """ Transfer table from stage to actuals (deprecated, use `transfer_actuals()`). """

    assert transfer_delta, '`transfer_delta == False` is no longer supported, use custom SQL-template to emulate this logic'

    return transfer_actuals(
        source_tables=source,
        target_tables=target,
        source_conn_id=source_conn_id,
        destination_conn_id=destination_conn_id,
        session=session,
        keep_temp_table=keep_temp_table,
        mode='sync' if as_ods else 'replace' if replace_data else 'update',
        **kwargs
    )

@deprecated('`transfer_actuals_tables()` is deprecated since 0.1.9, use `transfer_actuals()` instead')
def transfer_actuals_tables(
        source_tables: Iterable[str | Table],
        target_tables: Iterable[str | Table] | None = None,
        source_conn_id: str | None = None,
        destination_conn_id: str | None = None,
        group_id: str | None = None,
        num_parallel: int = 1,
        session: XComArg | ETLSession | None = None,
        transfer_delta: bool = True,
        as_ods: bool = False,
        keep_temp_table: bool = False,
        replace_data: bool = False,
        **kwargs) -> TaskGroup:

    """ Transfer tables from stage to actuals (deprecated, use `transfer_actuals()`). """

    assert transfer_delta, '`transfer_delta == False` is no longer supported, use custom SQL-template to emulate this logic'

    return transfer_actuals(
        source_tables=source_tables,
        target_tables=target_tables,
        source_conn_id=source_conn_id,
        destination_conn_id=destination_conn_id,
        group_id=group_id,
        num_parallel=num_parallel,
        session=session,
        keep_temp_table=keep_temp_table,
        mode='sync' if as_ods else 'replace' if replace_data else 'update',
        **kwargs
    )

def transfer_actuals(
        source_tables: Table | str | Iterable[str | Table],
        target_tables: Table | str | Iterable[str | Table] | None = None,
        source_conn_id: str | None = None,
        destination_conn_id: str | None = None,
        group_id: str | None = None,
        num_parallel: int = 1,
        session: XComArg | ETLSession | None = None,
        keep_temp_table: bool = False,
        mode: ActualsTransferMode = 'update',
        **kwargs) -> TaskGroup:

    """ Actuals (DDS) update """
    assert session is not None, 'Transfer to actuals requires ETL session'

    if isinstance(source_tables, (Table, str)):
        # Single table
        source_table = ensure_table(source_tables, source_conn_id)
        dest_table = ensure_table(target_tables or source_tables, destination_conn_id)

        op = ActualsTableTransfer(
            source_table=source_table,
            destination_table=dest_table,
            session=session,
            keep_temp_table=keep_temp_table,
            mode=mode,
            inlets=[source_table],
            outlets=[dest_table],
            **kwargs
        )
        return XComArg(op)

    # Multiple tables
    if target_tables and len(target_tables) != len(source_tables):
        raise AirflowFailException(f'Source and target tables list size must be equal')

    target_tables = target_tables or source_tables
    group_id = group_id or 'transfer-actuals'

    with TaskGroup(group_id, add_suffix_on_collision=True, prefix_group_id=False) as tg:
        ops_list = []
        for (source, target) in zip(source_tables, target_tables):
            source_table = ensure_table(source, source_conn_id)
            dest_table = ensure_table(target, destination_conn_id)
            op = ActualsTableTransfer(
                source_table=source_table,
                destination_table=dest_table,
                session=session,
                keep_temp_table=keep_temp_table,
                mode=mode,
                inlets=[source_table],
                outlets=[dest_table],
                **kwargs
            )
            ops_list.append(op)

        schedule_ops(ops_list, num_parallel)

    return tg

    # This function implements cross-database geterogenous data transfer.

    # It reads data from source table into memory and then sequentaly inserts 
    # each record into the target table. Fields order in the source and target tables 
    # must be identical and field types must be compatible, or transfer will fail or produce 
    # undesirable results.

    # To limit data selection or customize fields, a SQL template could be 
    # created for the source table (see `astro_extras.utils.template.get_template_file`).
    # The template must ensure fields order and type compatibility with the target table.
    # If transfer is running under `astro_extras.operators.session.ETLSession` context, 
    # a `session_id` field must also be manually added at proper place.

    # For example, if these tables are to participate in transfer:

    # ``` sql
    # create table source_data ( a int, b text );
    # create table target_data ( session_id int, b text, a int );
    # ```

    # then, this SQL template might be created as `source_data.sql` file:

    # ``` sql
    # select {{ti.xcom_pull(key="session").session_id}} as session_id, b, a 
    # from source_data;
    # ```

    # Args:
    #     source: Either a table name or a `Table` object which would be a data source.
    #         If a string name is provided, it may contain schema definition denoted by `.`. 
    #         For `Table` objects, schema must be defined in `Metadata` field,
    #         otherwise Astro SDK might fall to use its default schema.
    #         If a SQL template exists for this table name, it will be executed,
    #         otherwise all table data will be selected.

    #     target: Either a table name or a `Table` object where data will be saved into.
    #         If a name is provided, it may contain schema definition denoted by `.`. 
    #         For `Table` objects, schema must be defined in `Metadata` field,
    #         otherwise Astro SDK might fall to use its default schema.
    #         If omitted, `source` argument value is used (this makes sense only
    #         with string table name and different connections).

    #     source_conn_id: Source database Airflow connection.
    #         Used only with string source table name; for `Table` objects, `conn_id` field is used.
    #         If omitted and `session` argument is provided, `session.source_conn_id` will be used.

    #     destination_conn_id: Destination database Airflow connection.
    #         Used only with string target table name; for `Table` objects, `conn_id` field is used.
    #         If omitted and `session` argument is provided, `session.destination_conn_id` will be used.

    #     session:    `ETLSession` object. If set and no SQL template is defined,
    #         a `session_id` field will be automatically added to selection.

    #     changes_only:   If set to `True`, the operator will compare source and target
    #         tables and transfer data only when they are different. Target data are obtained
    #         by runnning `destination_sql`. By default, this query will be built using 
    #         `<destination_table>_a` view to get actual data.

    #         Deprecated since 0.1.1, use `transfer_changed_table` instead.

    #     kwargs:     Any parameters passed to underlying operator (e.g. `preoperator`, ...)

    # Returns:
    #     `XComArg` object

    # Examples:
    #     Using `Table` objects (note use of `Metadata` object to specify schemas):

    #     >>> with DAG(...) as dag:
    #     >>>     input_table = Table('table_data', conn_id='source_db', 
    #     >>>                          metadata=Metadata(schema='public'))
    #     >>>     output_table = Table('table_data', conn_id='target_db', 
    #     >>>                          metadata=Metadata(schema='stage'))
    #     >>>     transfer_table(input_table, output_table)

    #     Using string table name:

    #     >>> with DAG(...) as dag, ETLSession('source_db', 'target_db') as sess:
    #     >>>     transfer_table('public.table_data', session=sess)

    # """

def compare_table(
        source: str | BaseTable,
        target: str | BaseTable | None = None,
        source_conn_id: str | None = None,
        destination_conn_id: str | None = None,
        **kwargs) -> TaskGroup:

    """ Compare two tables.

    Creates a task of `CompareTableTransfer` operator to compare
    given tables, which would fail if any differences
    are detected among them.

    Args:
        source: Either a table name or a `Table` object which would be a data source.
            If a string name is provided, it may contain schema definition denoted by `.`. 
            For `Table` objects, schema must be defined in `Metadata` field,
            otherwise Astro SDK might fall to use its default schema.
            If a SQL template exists for this table name, it will be executed,
            otherwise all table data will be selected.

        target: Either a table name or a `Table` object where data will be saved into.
            If a name is provided, it may contain schema definition denoted by `.`. 
            For `Table` objects, schema must be defined in `Metadata` field,
            otherwise Astro SDK might fall to use its default schema.
            If omitted, `source` argument value is used (this makes sense only
            with string table name and different connections).

        source_conn_id: Source database Airflow connection.
            Used only with string source table name; for `Table` objects, `conn_id` field is used.
            If omitted and `session` argument is provided, `session.source_conn_id` will be used.

        destination_conn_id: Destination database Airflow connection.
            Used only with string target table name; for `Table` objects, `conn_id` field is used.
            If omitted and `session` argument is provided, `session.destination_conn_id` will be used.

    """
    source_table = ensure_table(source, source_conn_id)
    dest_table = ensure_table(target or source_table, destination_conn_id)

    op = CompareTableOperator(
        source_table=source_table, 
        destination_table=dest_table, 
        task_id=f'compare-{source_table.name}')
    return XComArg(op)

def compare_tables(
        source_tables: Iterable[str | Table],
        target_tables: Iterable[str | Table] | None = None,
        source_conn_id: str | None = None,
        destination_conn_id: str | None = None,
        group_id: str | None = None,
        num_parallel: int = 1,
        **kwargs) -> TaskGroup:

    """ Compares multiple tables.

    Creates an Airflow task group consisting of `CompareTableTransfer` operators 
    for each table pair from `source_tables` and `target_tables` lists.
    Each operator will compare source and target table and fails if any differences
    are detected among them.

    See `compare_table` for details.
    """
    if not target_tables or len(target_tables) != len(source_tables):
        raise AirflowFailException(f'Source and target tables list size must be equal')

    with TaskGroup(group_id or 'compare-tables', add_suffix_on_collision=True) as tg:
        ops_list = []
        for (source, target) in zip(source_tables, target_tables):
            source_table = ensure_table(source, source_conn_id)
            dest_table = ensure_table(target, destination_conn_id) or source_table
            op = CompareTableOperator(
                source_table=source_table, 
                destination_table=dest_table, 
                task_id=f'compare-{source_table.name}')
            ops_list.append(op)
        schedule_ops(ops_list, num_parallel)
    return tg

def update_timed_table(
        source: str | BaseTable,
        target: str | BaseTable | None = None,
        source_conn_id: str | None = None,
        destination_conn_id: str | None = None,
        session: XComArg | ETLSession | None = None,
        **kwargs) -> XComArg:
    
    """ Updates timed dictionary table (experimental) """

    source_table = ensure_table(source, source_conn_id)
    dest_table = ensure_table(target, destination_conn_id) or source_table
    op = TimedTableTransfer(
        source_table=source_table,
        destination_table=dest_table,
        session=session,
        **kwargs_with_datasets(kwargs=kwargs, 
                               input_datasets=source_table, 
                               output_datasets=dest_table)
    )
    return XComArg(op)

def update_timed_tables(
        source_tables: Iterable[str | Table],
        target_tables: Iterable[str | Table] | None = None,
        source_conn_id: str | None = None,
        destination_conn_id: str | None = None,
        group_id: str | None = None,
        num_parallel: int = 1,
        session: XComArg | ETLSession | None = None,
        **kwargs) -> TaskGroup:

    """ Updates multiple timed dictionary tables (experimental )"""

    if target_tables and len(target_tables) != len(source_tables):
        raise AirflowFailException(f'Source and target tables list size must be equal')

    target_tables = target_tables or source_tables

    with TaskGroup(group_id or 'transfer-tables', add_suffix_on_collision=True) as tg:
        ops_list = []
        for (source, target) in zip(source_tables, target_tables):
            source_table = ensure_table(source, source_conn_id)
            dest_table = ensure_table(target, destination_conn_id) or source_table
            op = TimedTableTransfer(
                source_table=source_table,
                destination_table=dest_table,
                session=session,
                **kwargs_with_datasets(
                    kwargs=kwargs, 
                    input_datasets=source_table, 
                    output_datasets=dest_table))
            ops_list.append(op)
        schedule_ops(ops_list, num_parallel)
    return tg




# Astro SDK Extras project
# (c) kol, 2023

""" FileChanged sensor """

import os
import asyncio
from datetime import datetime, timedelta
from functools import cached_property
from airflow.models import XCom
from airflow.configuration import conf
from airflow.hooks.filesystem import FSHook
from airflow.exceptions import AirflowException
from airflow.sensors.filesystem import FileSensor
from airflow.utils.context import Context
from airflow.triggers.base import BaseTrigger, TriggerEvent
from airflow.triggers.base import TriggerEvent
from typing import Tuple

class FileChangedTrigger(BaseTrigger):
    """
    A trigger that fires after it finds the requested file and this file has been changed since previous check.

    :param filepath: File name (relative to the base path set within the connection). Wildcards are not allowed.
    :param poll_interval: File check interval
    :param xcom_key: XCom key (``dag_id|task_id|exec_date``) to store file modification time (for internal use).
    """
    def __init__(
        self,
        filepath: str,
        poll_interval: float = 5.0,
        xcom_key: str = None,
    ):
        super().__init__()
        assert not any([s in filepath for s in ('*', '?')]), 'Wildcards are not supported'

        self.filepath = filepath
        self.poll_interval = poll_interval
        self.xcom_key = xcom_key
        self._file_key = None
        self._last_modified = None

        if xcom_key:
            self._dag_id, self._task_id, self._exec_date = xcom_key.split('|')
            self._exec_date = datetime.fromisoformat(self._exec_date)
            self._file_key = f'{os.path.split(filepath)[1]}_last_modified'

            xcom_val = XCom.get_one(
                execution_date=self._exec_date, 
                key=self._file_key, 
                task_id=self._task_id, 
                dag_id=self._dag_id, 
                include_prior_dates=True)
            if xcom_val:
                self._last_modified = datetime.fromisoformat(xcom_val)

            self.log.info(f'FileChangedTrigger started for xcom key {xcom_key}|{self._file_key}, last modified from xcom: {self._last_modified}')

    def serialize(self):
        return (
            "astro_extras.sensors.file.FileChangedTrigger",
            {
                "filepath": self.filepath,
                "poll_interval": self.poll_interval,
                "xcom_key": self.xcom_key,
            },
        )

    async def run(self):
        while True:
            if os.path.exists(self.filepath):
                mod_time = datetime.fromtimestamp(os.path.getmtime(self.filepath))
                self.log.info("Found file %s, last modified: %s", self.filepath, mod_time.isoformat())
                if not self._file_key:
                    yield TriggerEvent((self.filepath, mod_time))
                else:
                    if self._last_modified and mod_time <= self._last_modified:
                        self.log.info("File %s was not changed since last check", self.filepath)
                    else:
                        self.log.info("Triggering file %s change event", self.filepath)
                        self._last_modified = mod_time
                        XCom.set(
                            key=self._file_key, 
                            value=mod_time.isoformat(), 
                            task_id=self._task_id, 
                            dag_id=self._dag_id, 
                            execution_date=self._exec_date)
                        yield TriggerEvent((self.filepath, mod_time))

            await asyncio.sleep(self.poll_interval)

class FileChangedSensor(FileSensor):
    """
    Extension of standard ``FileSensor`` which waits for a file to appear on a filesystem,
    and to be modified since last check.

    Args:
        fs_conn_id: reference to the File (path) connection id
        filepath: file name to look after (wildcards/directories not allowed)
        deferrable: triggers Airflow task deferral mode

    Returns:
        2-element tuple with file name and file modification time as isoformatted string
    """

    def __init__(
            self,
            *,
            deferrable: bool = conf.getboolean("operators", "default_deferrable", fallback=False),
            **kwargs,
        ):
            super().__init__(**kwargs)
            self.deferrable = deferrable

    @cached_property
    def path(self) -> str:
        hook = FSHook(self.fs_conn_id)
        basepath = hook.get_path()
        full_path = os.path.join(basepath, self.filepath)
        return full_path

    def execute(self, context: Context) -> None:
        if not self.deferrable:
            super().execute(context=context)
        if not self.poke(context=context):
            self.defer(
                timeout=timedelta(seconds=self.timeout),
                trigger=FileChangedTrigger(
                    filepath=self.path,
                    poll_interval=self.poke_interval,
                    xcom_key=f'{self.dag_id}|{self.task_id}|{context["ti"].execution_date.isoformat()}'
                ),
                method_name="execute_complete",
            )

    def execute_complete(self, context: Context, event: Tuple[str, datetime]) -> Tuple[str, str]:
        if not event:
            raise AirflowException("%s task failed as %s not found", self.task_id, self.filepath)
        self.log.info("%s completed successfully, file %s modified at %s found", self.task_id, event[0], event[1])
        return event[0], event[1].isoformat()



# Astro SDK Extras project
# (c) kol, 2023

from airflow.sensors.external_task import ExternalTaskSensor

class AutoOnlyTaskSensor(ExternalTaskSensor):
    """ A DAG communication operator that takes into account the way DAG is launched.

    If the pending DAG was started manually, the operator will not wait for the expected DAG to complete
    and the execution will continue immediately.
    """

    def execute(self, context):
        """ Primary operator execution method """
        if context['run_id'].find('manual') == 0:
            self.log.info('Manual execution detected, skipping DAG wait')
        else:
            super().execute(context)
